!   Michael S. Briggs,  Michael.Briggs@uah.edu.
!   University of Alabama in Huntsville.
!   copyright 2013 & 2014 & 2016 Jan 5.

module MSB_Constants


#ifdef __GFORTRAN__

   !  Fixes a gfortran bug, in which the value of real128 is 10 instead of 16.
   !  See https://gcc.gnu.org/bugzilla/show_bug.cgi?id=52153
   !  We test for the compiler and only replace the value of "real128" for
   !  gfortran.   The value of real128 is a compiler choice and we don't know
   !  the correct value for other compilers.  It's 16 for most compilers,
   !  but might be 4, or ...

   use, intrinsic :: ISO_FORTRAN_ENV,  real128_BUG_VALUE => real128

   implicit none

   integer, parameter :: real128 = 16

#else

   use, intrinsic :: ISO_FORTRAN_ENV

   implicit none

#endif


   real (real128), parameter :: MJD_OFFSET = 2400000.5_real128

   real (real32), parameter :: PI_SNGL = 3.1415926_real32
   real (real64), parameter :: PI_DBL = 3.1415926535897932_real64
   real (real128), parameter :: PI_QUAD = 3.14159265358979323846264338327950288_real128

   real (real32), parameter :: DEGREES_TO_RADIANS_SNGL = PI_SNGL / 180.0_real32
   real (real64), parameter :: DEGREES_TO_RADIANS_DBL = PI_DBL / 180.0_real64
   real (real128), parameter :: DEGREES_TO_RADIANS_QUAD = PI_QUAD / 180.0_real128

   real (real32), parameter :: RADIANS_TO_DEGREES_SNGL = 180.0_real32 / PI_SNGL
   real (real64), parameter :: RADIANS_TO_DEGREES_DBL = 180.0_real64 / PI_DBL
   real (real128), parameter :: RADIANS_TO_DEGREES_QUAD = 180.0_real128 / PI_QUAD


   !  From http://en.wikipedia.org/wiki/Earth_radius.
   !  Mean radius defined by the International Union of Geodesy and Geophysics as a weighted
   !  mean of the Equatorial and Polar radii.

   real (real32), parameter :: EARTH_MEAN_RADIUS_M_SNGL = 6371009.0_real32    ! m
   real (real64), parameter :: EARTH_MEAN_RADIUS_M_DBL = 6371009.0_real64    ! m


   !  speed of light, m / s

   real (real64), parameter :: LIGHT_SPEED_MPS_DBL = 299792458.0_real64



end module MSB_Constants
