!  Copyright Michael S. Briggs, University of Alabama in Huntsville.
!  For NASA's Fermi Gamma-ray Burst Monitor (GBM).
!  2008 -- 2016.

! #################################################################################################


module some_gbm_constants

   integer, parameter :: NUM_SPEC_CHAN = 128

   integer, parameter :: FILE_NAME_MAX_LEN = 200

end module some_gbm_constants

module Read_Lvl1_TTE_Interface



   interface

      !  Fortran 95 interface to the C-language routine Read_Lvl1_TTE, which uses CFITSIO to read
      !  raw / event-list TTE from GBM Level 1 FITS files.

      subroutine Read_Lvl1_TTE      &
           ( InputFileName, DetNum,      &
             FileStartTime, FileStopTime, TriggerTime, Num_TTE, FITS_DataSum,  &
             tte_chan_Cptr, tte_time_Cptr, EMin_of_chan, EMax_of_chan )        &
            bind ( C, name='Read_Lvl1_TTE' )

         use, intrinsic :: iso_c_binding

         use some_gbm_constants

         implicit none

         ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

         character (kind=C_CHAR, len=1), dimension (FILE_NAME_MAX_LEN), intent (in) :: InputFileName
         integer (C_LONG), intent (in), VALUE :: DetNum

         real (C_DOUBLE), intent (out) :: FileStartTime, FileStopTime   ! time range of file.  FITS keywords TSTART & TSTOP

         !  TriggerTime is only valid for trigger events, in which case the value returned will
         !  be larger than 237427201.0 (2008 July 11, the day of GBM trigger enable).
         !  If the data file is for continuous (i.e., non-trigger) data, then -99.99 is returned.
         real (C_DOUBLE), intent (out) :: TriggerTime

         integer (C_LONG), intent (out) :: Num_TTE
         character (kind=C_CHAR, len=1), dimension (11), intent (out) :: FITS_DataSum   !  FITS keyword DATASUM of the EVENT header
         type (C_PTR), intent (out) :: tte_chan_Cptr
         type (C_PTR), intent (out) :: tte_time_Cptr
         real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1), intent (out) :: EMin_of_chan
         real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1), intent (out) :: EMax_of_chan

      end subroutine Read_Lvl1_TTE


      ! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^

      !  Revision 2014 February 11: the C routine no longer deallocates the FITS file.
      !  That is now automatically done by C-routine Read_Lvl1_TTE.   This supersedes
      !  that aspect of the comments in the next paragraph.

      !  Fortran 95 interface to the C-language routine Cleanup_TTE_FitsFile.
      !  Cleanup_TTE_FitsFile can (should) be called after calling Read_Lvl1_TTE -- it
      !  deallocates one FITS file and two malloc'd arrays.
      !  It can be called after additional calls to Read_Lvl1_TTE have been made, but must
      !  be called before a second call to Read_Lvl1_TTE is made for the same detector --
      !  the two routines communicate "on the side" and can only remember one set of "file and
      !  two arrays per detector".   If Read_Lvl1_TTE is called twice for one particular detector
      !  with no intervening call to Cleanup_TTE_FitsFile, there will be lost, unfreed memory.


      subroutine Cleanup_TTE_FitsFile ( DetNum )  bind ( C, name="Cleanup_TTE_FitsFile" )

         use, intrinsic :: iso_c_binding

         ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

         integer (C_LONG), intent (in), value :: DetNum

      end subroutine Cleanup_TTE_FitsFile



      !  For debugging.  Not needed in normal use of the code.

      subroutine ReportSharedVar ( DetNum )  bind ( C, name="ReportSharedVar" )

         use, intrinsic :: iso_c_binding

         ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

         integer (C_LONG), intent (in), value :: DetNum

      end subroutine ReportSharedVar

   end interface


end module Read_Lvl1_TTE_Interface


! #################################################################################################
