
!  Michael S. Briggs, UAHuntsville & NSSTC, copyright 2009 February 9.
!  First release: version 0.9
!  Version 0.9.1, 2009 Feb 12: add PNG output option.
!  Version 0.9.5, 2009 March 1: upgraded subroutine Read_Lvl1_TTE to read energy edges.
!  Version 0.9.6, 2009 March 7: moved all user input from Make_TTE_OneHisto to
!    the main routine so that Make_TTE_OneHisto can be called by pipeline programs
!    that don't have an interactive user.
!  Revision 0.99 by MSB, 2009 March 22: call widget routine for energy channels,
!     include energy range in channel label when possible.
!  Revision 1.0 by MSB, 2009 May 25: optional auto-selection of axis limits and tick spacing.
!  Revision 1.1 by MSB, 2009 May 31
!  Revision 1.1.1 by MSB, 2009 June 1: add option to change start of Y-axis.
!  Revision 1.2 by MSB, 2009 June 21: option to make right Y-axis be count rate either per
!  detector, or summed over detectors.
!  Revision 1.3 by MSB, 2009 Aug 15: 1) add eps file format option in this file, 2) thicker lines
!  and larger characters in subroutine Make_TTE_OneHisto, 3) Output filename can be input by
!  user, or can use default constructed from trigger bn-number.
!  Revision 1.4 by MSB, 2011 Feb 14: supports correcting the times for the drift of the GBM
!  DPU clock.

!  Program TTE_Histo_OneGraph: plots histograms of GBM TTE data.  Black curve is observed (raw)
!  counts, optional red histogram is deadtime corrected.
!  One plot is made, which sums from 1 to all 14 detectors.   "Summing 1" detectors displays
!  just one detector.

!  Usage: TTE_Histo_OneGraph.exe  bnyymmhhfff
!  OR
!  Usage: TTE_Histo_OneGraph.exe  bnyymmhhfff   VerNum
!  "yymmhhfff" is the trigger number, VerNum is the file version number, which may be omitted if 00.


! #################################################################################################


program TTE_Histo_OneGraph

   use MSB_Constants

   use, intrinsic :: iso_c_binding, only: C_CHAR, C_DOUBLE

   use TTE_subroutines, only:  TTE_GetCmdLineInput_AllFileNames,  MakeEnergyLabels,                    &
             Read_and_Bin_TTE, Select_Detectors,   Make_TTE_OneHisto,                                  &
             auto_xdiv_value, auto_ymax_value, auto_ydiv_value, Get_OutputFileName, Get_GraphFormat,   &
             MakeUpperCase, GapTimeQuestion, TTE_OneHisto_2_file

   use TTE_constant_defs
   use TTE_Histogram_Arrays    !  module with global arrays

   implicit none

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=FILE_NAME_MAX_LEN) :: NamePortion
   character (kind=C_CHAR, len=FILE_NAME_MAX_LEN), dimension(0: MAX_GBM_DET-1 ) :: InputFileNames_for_C

   character (len=6), parameter :: Fragment = ".Histo"
   character (len=FILE_NAME_MAX_LEN) :: OutputFileName

   real (real64), dimension (2) :: GraphTimeRange_s
   real (real64) :: BinWidth_s
   integer (int32), dimension (2) :: ChannelRange

   real (real32), dimension ( 0: NUM_SPEC_CHAN-1, 0: MAX_GBM_DET-1 ) :: EMin_of_chan
   real (real32), dimension ( 0: NUM_SPEC_CHAN-1, 0: MAX_GBM_DET-1 ) :: EMax_of_chan

   integer (int32) :: idet
   integer :: TimeMode

   character (len=2) :: TimeLabel, BinLabel

   logical, dimension ( 0: MAX_GBM_DET-1 ) :: RequestDetectors
   integer (int32) :: NumRequestDetectors
   integer (int32) :: SingleRequestDetector
   logical :: Single_DetType_Selected
   character (len=DETECTOR_SELECTION_LABEL_LEN) :: DetectorSelectionLabel
   character (len=42) :: ChannelRangeLabel

   character (len=40) :: input
   real (real32) :: try
   real (real32) :: TimeRange
   real (real32) :: xdiv
   real (real32) :: ydiv
   real (real32) :: ymin
   real (real32) :: ymax
   character (len=1) :: PlotDeadtimeCorr
   character (len=3) :: GraphFormat

   integer (int32) :: MinBinValue, MaxBinValue
   real (real32) :: Corrected_Min_BinValue, Corrected_Max_BinValue

   character (len=ENERGY_LABEL_LEN) :: Left_Energy_Label
   character (len=ENERGY_LABEL_LEN) :: Right_Energy_Label

   logical, parameter :: Determine_Channel_Range = .TRUE.
   integer (int32) :: FirstDetSelected = 0     ! initial value to suppress warning message

   logical :: RightAxis_is_PerDet

   real (real32) :: GapTime
   logical :: ClockDrift_Flag

   !  This is the TriggerTime, read from the FITS file, for trigger data.
   !  For continuous / non-trigger data, it is a user input value.
   real (kind=real128) :: ReferenceTime

   character (len=9) :: TimeRefType


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   write ( *, '( / "TTE_Histo_OneGraph by MSB, Version 1.5, 2011 March 6." / )' )


   !  read command line input, obtain name of datafiles to read:

   call TTE_GetCmdLineInput_AllFileNames ( NamePortion, InputFileNames_for_C )



   !  more input from user:

   !  Primary output, from user input: detector selection.  Secondary output: derived quantities from detector selection:

   call Select_Detectors     &
         ( RequestDetectors, NumRequestDetectors, SingleRequestDetector, Single_DetType_Selected, DetectorSelectionLabel )


   !  Ask for time "mode": units of X-axis and binning:

   1010 continue
   write ( *, '( / "There are three modes for the units for inputting time selection and labeling the plot:" )' )
   write ( *, ' ( "     x-axis   bin widths" / "1:     s        ms" / "2:    ms        ms" / "3:    ms       microsec" )' )
   write ( *, '( "Select input/label mode: " )', advance='no' )
   read ( *, *, ERR=1010 )  TimeMode
   if ( TimeMode /= 1  .and.  TimeMode /= 2  .and.  TimeMode /= 3 )  goto 1010

   if ( TimeMode == 1 ) then
      TimeLabel = "s "
   else
      TimeLabel = "ms"
   end if
   if ( TimeMode /= 3 ) then
      BinLabel = "ms"
   else
      BinLabel = "us"
   end if


   !  Ask for time range of X-axis:

   1020 continue
   write ( *, '( / "Input Time Range to plot for file ", A, " in ", A, ": " )', advance='no' )  trim (NamePortion), TimeLabel
   read ( *, *, err=1020 )  GraphTimeRange_s (1), GraphTimeRange_s (2)

   TimeRange = real (  GraphTimeRange_s (2) - GraphTimeRange_s (1),  real32 )    ! units are ms or s, depending on TimeLabel


   !  If user input time range in ms, convert to seconds, the units of GraphTimeRange_s:
   if ( TimeLabel == "ms" )  GraphTimeRange_s = GraphTimeRange_s / 1000.0_real64


   write ( *, '( "The time range of the graph will be", F8.3, " to", F8.3, " s." )' )  GraphTimeRange_s (1), GraphTimeRange_s (2)
   write ( *, '( "or", F8.1, " to", F8.1, " ms." )' )    &
          GraphTimeRange_s (1) * 1000.0_real64, GraphTimeRange_s (2) * 1000.0_real64


   !  Determine tick-spacing of X-axis, either automatically (with user approval), or from user input:

   try = auto_xdiv_value ( TimeRange )

   9020 continue
   write ( *, 9030, advance='no' )  TimeLabel, try, TimeLabel
   9030 format ( / "Input X (time) axis major (numbered) tick spacing in ", A, " or ENTER to use", F8.2, 1X, A, " : " )
   read ( *, '(A)', err=9020 ) input

   if ( len_trim (input) == 0 ) then
      xdiv = try
   else
      read ( input, *, err=9020 )  xdiv
   end if


   !  Ask for histogram bin width:

   1030 continue
   write ( *, '( / "Input Histogram Bin Width in ", A, ": " )', advance='no' )  BinLabel
   read ( *, *, err=1030 )  BinWidth_s

   !  BinWidth_s variable has units of s, but user input in
   !  either ms or us -- convert from their input to seconds.
   if ( BinLabel == "ms" )  then
      BinWidth_s = BinWidth_s / 1000.0_real64
   else
      BinWidth_s = BinWidth_s / 1.0E+6_real64
   end if

   write ( 6, '( / "BinWidth in s:", 1PG13.5 )' ) BinWidth_s


   call GapTimeQuestion  ( ClockDrift_Flag, GapTime )


   !  Read the TTE data from the Level 1 Fits file, then bin it into the shared arrays of
   !  the module TTE_Histogram_Arrays.

   call Read_and_Bin_TTE  ( InputFileNames_for_C, GraphTimeRange_s, BinWidth_s, RequestDetectors,      &
            Single_DetType_Selected, Determine_Channel_Range, GapTime, ClockDrift_Flag,                &
            ChannelRange, EMin_of_chan, EMax_of_chan, ReferenceTime, TimeRefType )



   !  More user input:


   !  We base the energy boundaries of the channel on one particular detector -- the first requested one.
   !  (The user also made their channel selections based on the energy edges of this detector.)
   !  We could be more clever and average across the selected detectors, at least if they were a common type.   ????

   !  Figure out the detector number of the first detector on the list of selected detectors.

   FindFirstDet: do idet=0, MAX_GBM_DET - 1

      if ( RequestDetectors (idet) ) then
         FirstDetSelected = idet
         exit FindFirstDet
      end if

   end do FindFirstDet



   write ( ChannelRangeLabel, '( "Channels ", I3, " to ", I3 )' )  ChannelRange



   !  If a single detector type was selected so that the channels have a common meaning in energy,
   !  create "nice" labels for first and last selected energy values.

   if ( Single_DetType_Selected ) then

      call MakeEnergyLabels  ( EMin_of_chan ( :, FirstDetSelected ), EMax_of_chan ( :, FirstDetSelected ),      &
               ChannelRange (1), ChannelRange (2), Left_Energy_Label, Right_Energy_Label )

      ChannelRangeLabel =   &
      ChannelRangeLabel (1:19) // " = " // trim(Left_Energy_Label) // " to " // trim(Right_Energy_Label)

   end if    ! Single_DetType_Selected ?



   !   Ask whether to plot deadtime corrected histogram ?

   1050 continue
   write ( *, '( // "Do you want to overplot with a deadtime corrected histogram?" / "Y, N or ENTER for N: " )', advance='no' )
   read ( *, '(A)', err=1050 )  PlotDeadtimeCorr
   if ( len_trim (PlotDeadtimeCorr) == 0 )  PlotDeadtimeCorr = 'N'
   call MakeUpperCase ( PlotDeadtimeCorr )


   !  Ask about y-max value for y-axis -- first provide some information to the user:

   !  Two arrays obtained from module TTE_Histogram_Arrays.
   !  (Deadtime corrected min & max may not be in the same bin as the non-deadtime corrected value...)

   !  min & max values over the bins of the histogram of the raw data, histogram of sum of all detectors:
   MinBinValue = minval ( GBL_Data_Histogram_SummedDets )
   MaxBinValue = maxval ( GBL_Data_Histogram_SummedDets )

   !  min & max values, this time deadtime corrected:
   Corrected_Min_BinValue = minval ( GBL_Deadtime_Histogram_SummedDets )
   Corrected_Max_BinValue = maxval ( GBL_Deadtime_Histogram_SummedDets )


   if ( PlotDeadtimeCorr == 'Y' ) then
      write ( *, 8010 )  MaxBinValue, Corrected_Max_BinValue, MinBinValue, Corrected_Min_BinValue
      8010 format ( / "Maximum and minimum values observed in a histogram bin:" /  &
                      "     Observed    Deadtime corrected" /                      &
                      " Max: ", I5, 9X, F8.1 /                                     &
                      " Min: ", I5, 9X, F8.1 / )
      try = auto_ymax_value  ( Corrected_Max_BinValue )
   else
      write ( *, 8020 )  MaxBinValue, MinBinValue
      8020 format ( / "Maximum and minimum values observed in a histogram bin:" /  &
                      "     Observed" /                                            &
                      " Max: ", I5 /                                               &
                      " Min: ", I5 / )
      try = auto_ymax_value  ( real (MaxBinValue, real32) )
   end if


   !  Now determine the minimum and maximum values of the y-axis, either automatically (with user approval),
   !  or from user input.
   !  Same for tick spacing of y-axis.

   1055 continue
   write ( *, '( / "Input the value to use for the BOTTOM of the Y-axis or ENTER to use 0.0: " )', advance='no' )
   read ( *, '(A)' )  input
   if ( len_trim (input) == 0 ) then
      ymin = 0.0
   else
      read ( input, *, err=1055 )  ymin
   end if

   1060 continue
   write ( *, '( / "Input the value to use for the TOP of the Y-axis or ENTER to use", F9.1, ": " )', advance='no' )  try
   read ( *, '(A)' )  input
   if ( len_trim (input) == 0 ) then
      ymax = try
   else
      read ( input, *, err=1060 )  ymax
   end if

   try = auto_ydiv_value  ( ymax )
   write ( *, '( "Input the Y-axis major (numbered) tick spacing or ENTER to use", F7.1, ": " )', advance='no' )  try
   read ( *, '(A)' )  input
   if ( len_trim (input) == 0 ) then
      ydiv = try
   else
      read ( input, *, err=1060 )  ydiv
   end if


   if ( NumRequestDetectors > 1 ) then
      1070 continue
      write ( *, '( / "Do you want the right axis for count rate to be PER DETECTOR or for SUMMED over detectors?" )' )
      write ( *, '( "Input ''P'' or ENTER for PER Detector or ''S'' for SUMMED over detectors: " )', advance='no' )
      read ( *, '(A)', err=1070 )  input
      if ( len_trim (input) == 0 )  input = 'P'
      if ( input /= 'P' .AND. input /= 'p' .AND. input /= 'S' .AND. input /= 's' ) goto 1070
      RightAxis_is_PerDet = .FALSE.
      if ( input == 'P' .OR. input == 'p' )  RightAxis_is_PerDet = .TRUE.
   else
      RightAxis_is_PerDet = .FALSE.
   end if


   call  Get_GraphFormat  ( GraphFormat )

   call Get_OutputFileName  ( NamePortion, Fragment, GraphFormat, OutputFileName )



   !  Plot a Histogram of the TTE events !!!
   !  Data array obtained from module TTE_Histogram_Arrays.
   !  Summed array contains the requested detectors, from a single detector to all 14.

   call Make_TTE_OneHisto  ( GBL_Data_Histogram_SummedDets, GBL_Deadtime_Histogram_SummedDets, GraphTimeRange_s,            &
           BinWidth_s, NumRequestDetectors, OutputFileName, NamePortion, TimeLabel, BinLabel, DetectorSelectionLabel,       &
           ChannelRangeLabel, xdiv, ydiv, ymin, ymax, PlotDeadtimeCorr, GraphFormat, RightAxis_is_PerDet,                   &
           ClockDrift_Flag, GapTime, ReferenceTime, TimeRefType )


   ! output the histogram to a text file
#ifdef TEXTOUT
   call TTE_OneHisto_2_file ( GBL_Data_Histogram_SummedDets, GraphTimeRange_s, BinWidth_s, NamePortion )
#endif



   !  Cleanup
   !  Deallocate global arrays allocated in Fortran, the arrays of module TTE_Histogram_Arrays.
   !  (Not really necessary since this program is about to exit, but might be useful in
   !   other programs.)

   deallocate (GBL_Data_Histogram_SummedDets)
   deallocate (GBL_Deadtime_Histogram_SummedDets)
   deallocate (GBL_Data_Histogram_ByDet)
   deallocate (GBL_Deadtime_Histogram_ByDet)


   stop

end program TTE_Histo_OneGraph


! #################################################################################################

