!  Copyright Michael S. Briggs, University of Alabama in Huntsville.
!  For NASA's Fermi Gamma-ray Burst Monitor (GBM).
!  copyright 2009 August 14, rev. 2009 Nov 26.
!  rev. 2011 June 24.  rev 2016 Jan 27.


!  Program to convert GLAST/Fermi Mission Elapsed Time (MET) to calendar date and time
!  (hours, minutes, seconds) UT.   The MET must be in units of seconds, rather than raw
!  GBM values of tenths of seconds, 2 microseconds.


program Call_MET_to_UTC

   use MSB_Constants

   use Time_and_Coord_Routines

   implicit none


   !  MET time has 9 digits before the decimal point, 6 after (to microseconds),
   !  for a total of 15 decimal digits.  This just matches double-precision, but I
   !  choose not to use double precision to have margin, using quad-precision
   !  (real128) instead.

   real (real128) :: GBM_MET

   real (real128) :: JulianDay

   integer (int32) :: year
   integer (int32) :: month
   integer (int32) :: day
   integer (int32) :: hours
   integer (int32) :: minutes
   real (real64) :: seconds

   integer (int32) :: seconds_INT
   real (real64) :: sub_seconds
   character (len=9) :: seconds_string

   integer (int32) :: NumberCmdArguments
   character (len=100) :: string


   ! * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *


   write ( *, '( / "MET_to_UTC by MSB, version 1.2, 2014 April 8." )' )


   !  get input from the command line:

   NumberCmdArguments = command_argument_count ()
   if ( NumberCmdArguments /= 1 ) then
      write ( *, '( / "Usage: ./MET_to_UTC  met_value" /)' )
      stop
   end if

   call get_command_argument  ( 1, string )
   read ( string, * )  GBM_MET

   write ( *, '( / "Received MET time:", F17.6, " s" )' )  GBM_MET


   call MET_to_UTC ( GBM_MET, JulianDay, year, month, day, hours, minutes, seconds, seconds_string )


   write ( *, '( / I4, "-", I2.2, "-", I2.2, 2X, I2.2, ":", I2.2, ":", A9  )' )     &
       year, month, day, hours, minutes, seconds_string


   stop


end program Call_MET_to_UTC
