!  Copyright Michael S. Briggs, University of Alabama in Huntsville.
!  For NASA's Fermi Gamma-ray Burst Monitor (GBM).
!   copyright 2010.  revised 2016 Jan 27.

!   Modified by MSB 2011 Feb 9 to add Quaternion and Angular Velocity.
!   Modified by MSB 2013 Dec 1, adding output of Position vector.

module Read_Lvl1_SC_PosHist_mod

   use, intrinsic :: iso_c_binding, only: c_char, c_float, c_double, c_long, c_ptr, c_f_pointer

   implicit none


   PRIVATE

   PUBLIC :: FILE_NAME_MAX_LEN, Read_Lvl1_SC_PosHist_SV, Read_Lvl1_SC_PosHist_Fort,    &
       Cleanup_Read_Lvl1_SC_PosHist_Fort, Interpolate_Convert_SC_PosHist


   integer, parameter :: FILE_NAME_MAX_LEN = 200


   interface Read_Lvl1_SC_PosHist_DEF


      ! . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .


      !  This version returns selected values, from the records just before & after the
      !  requested time.   This version can be directly called from Fortran "user" code.

      subroutine Read_Lvl1_SC_PosHist_SV  ( InputFileName, Request_MET, Before_MET, After_MET,      &
              Before_Quaternion, After_Quaternion, Before_AngVelocity, After_AngVelocity,           &
              Before_Position, After_Position, Before_Velocity, After_Velocity )                    &
              bind ( C, name = "Read_Lvl1_SC_PosHist_SV" )

         import

         character (kind=C_CHAR, len=1), dimension (FILE_NAME_MAX_LEN), intent (in) :: InputFileName
         real (c_double), intent (in), VALUE :: Request_MET

         real (c_double), intent (out) :: Before_MET
         real (c_double), intent (out) :: After_MET

         real (c_double), dimension (4), intent (out) :: Before_Quaternion
         real (c_double), dimension (4), intent (out) :: After_Quaternion

         real (c_double), dimension (3), intent (out) :: Before_AngVelocity
         real (c_double), dimension (3), intent (out) :: After_AngVelocity

         real (c_float), dimension (3), intent (out) :: Before_Position
         real (c_float), dimension (3), intent (out) :: After_Position

         real (c_float), dimension (3), intent (out) :: Before_Velocity
         real (c_float), dimension (3), intent (out) :: After_Velocity

      end subroutine Read_Lvl1_SC_PosHist_SV


      ! . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

      !  This version returns all values contained in the file.
      !  Probably one will want to call the provided Fortran-interface subroutine
      !  "Read_Lvl1_SC_PosHist_Fort", rather than calling this routine directly.

      subroutine Read_Lvl1_SC_PosHist  ( InputFileName, Num_History_Recs, METs_of_Records,     &
                  Quaternion_1_vs_t, Quaternion_2_vs_t, Quaternion_3_vs_t, Quaternion_4_vs_t,  &
                  AngVelocity_1_vs_t, AngVelocity_2_vs_t, AngVelocity_3_vs_t,                  &
                  X_position_vs_t, Y_position_vs_t, Z_position_vs_t,                           &
                  X_velocity_vs_t, Y_velocity_vs_t, Z_velocity_vs_t )                          &
              bind ( C, name = "Read_Lvl1_SC_PosHist" )

         import

         character (kind=C_CHAR, len=1), dimension (FILE_NAME_MAX_LEN), intent (in) :: InputFileName

         integer (C_LONG), intent (out) :: Num_History_Recs

         type (C_PTR), intent (out) :: METs_of_Records

         type (C_PTR), intent (out) :: Quaternion_1_vs_t
         type (C_PTR), intent (out) :: Quaternion_2_vs_t
         type (C_PTR), intent (out) :: Quaternion_3_vs_t
         type (C_PTR), intent (out) :: Quaternion_4_vs_t

         type (C_PTR), intent (out) :: AngVelocity_1_vs_t
         type (C_PTR), intent (out) :: AngVelocity_2_vs_t
         type (C_PTR), intent (out) :: AngVelocity_3_vs_t

         type (C_PTR), intent (out) :: X_position_vs_t
         type (C_PTR), intent (out) :: Y_position_vs_t
         type (C_PTR), intent (out) :: Z_position_vs_t

         type (C_PTR), intent (out) :: X_velocity_vs_t
         type (C_PTR), intent (out) :: Y_velocity_vs_t
         type (C_PTR), intent (out) :: Z_velocity_vs_t

      end subroutine Read_Lvl1_SC_PosHist


      ! . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

      !  Call this routine when the vectors are no longer needed, in order
      !  to free the memory.
      !  Probably one will want to call the provided Fortran-interface subroutine
      !  "Cleanup_Read_Lvl1_SC_PosHist_Fort", rather than calling this routine directly.

      subroutine Cleanup_Read_Lvl1_SC_PosHist  ( METs_of_Records,         &
                   Quaternion_1_vs_t, Quaternion_2_vs_t, Quaternion_3_vs_t, Quaternion_4_vs_t,  &
                   AngVelocity_1_vs_t, AngVelocity_2_vs_t, AngVelocity_3_vs_t,                  &
                   X_position_vs_t, Y_position_vs_t, Z_position_vs_t,     &
                   X_velocity_vs_t, Y_velocity_vs_t, Z_velocity_vs_t )    &
              bind ( C, name = "Cleanup_Read_Lvl1_SC_PosHist" )

         import

         type (C_PTR), intent (out) :: METs_of_Records

         type (C_PTR), intent (out) :: Quaternion_1_vs_t
         type (C_PTR), intent (out) :: Quaternion_2_vs_t
         type (C_PTR), intent (out) :: Quaternion_3_vs_t
         type (C_PTR), intent (out) :: Quaternion_4_vs_t

         type (C_PTR), intent (out) :: AngVelocity_1_vs_t
         type (C_PTR), intent (out) :: AngVelocity_2_vs_t
         type (C_PTR), intent (out) :: AngVelocity_3_vs_t

         type (C_PTR), intent (out) :: X_position_vs_t
         type (C_PTR), intent (out) :: Y_position_vs_t
         type (C_PTR), intent (out) :: Z_position_vs_t

         type (C_PTR), intent (out) :: X_velocity_vs_t
         type (C_PTR), intent (out) :: Y_velocity_vs_t
         type (C_PTR), intent (out) :: Z_velocity_vs_t

      end subroutine Cleanup_Read_Lvl1_SC_PosHist


   end interface Read_Lvl1_SC_PosHist_DEF



   !  These shared variables are PRIVATE to this file.  They are shared between
   !  Read_Lvl1_SC_PosHist_Fort and Cleanup_Read_Lvl1_SC_PosHist_Fort.
   !  Fortran Read_Lvl1_SC_PosHist_Fort calls C routine Read_Lvl1_SC_PosHist which
   !  calls malloc to connect memory to these pointers.  Fortran routine
   !  Cleanup_Read_Lvl1_SC_PosHist_Fort calls C routine Cleanup_Read_Lvl1_SC_PosHist
   !  which calls free to deallocate the memory.   The shared variable design is
   !  used so that these variables don't have to be arguments to Read_Lvl1_SC_PosHist_Fort
   !  and Cleanup_Read_Lvl1_SC_PosHist_Fort -- the calling program doesn't need to
   !  "know" about them.

   type (C_PTR) :: METs_of_Records_Cptr

   type (C_PTR) :: Quaternion_1_vs_t_Cptr
   type (C_PTR) :: Quaternion_2_vs_t_Cptr
   type (C_PTR) :: Quaternion_3_vs_t_Cptr
   type (C_PTR) :: Quaternion_4_vs_t_Cptr

   type (C_PTR) :: AngVelocity_1_vs_t_Cptr
   type (C_PTR) :: AngVelocity_2_vs_t_Cptr
   type (C_PTR) :: AngVelocity_3_vs_t_Cptr

   type (C_PTR) :: X_position_vs_t_Cptr
   type (C_PTR) :: Y_position_vs_t_Cptr
   type (C_PTR) :: Z_position_vs_t_Cptr

   type (C_PTR) :: X_velocity_vs_t_Cptr
   type (C_PTR) :: Y_velocity_vs_t_Cptr
   type (C_PTR) :: Z_velocity_vs_t_Cptr


contains


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


   subroutine Read_Lvl1_SC_PosHist_Fort  ( InputFileName, Num_History_Recs, METs_of_Records,  &
        Quaternion_1_vs_t, Quaternion_2_vs_t, Quaternion_3_vs_t, Quaternion_4_vs_t,           &
        AngVelocity_1_vs_t, AngVelocity_2_vs_t, AngVelocity_3_vs_t,                           &
        X_position_vs_t, Y_position_vs_t, Z_position_vs_t,                                    &
        X_velocity_vs_t, Y_velocity_vs_t, Z_velocity_vs_t  )


      character (kind=c_char, len=FILE_NAME_MAX_LEN), intent (in) :: InputFileName

      integer (C_LONG), intent (out) :: Num_History_Recs

      real (C_DOUBLE), dimension (:), pointer :: METs_of_Records

      real (C_DOUBLE), dimension (:), pointer :: Quaternion_1_vs_t
      real (C_DOUBLE), dimension (:), pointer :: Quaternion_2_vs_t
      real (C_DOUBLE), dimension (:), pointer :: Quaternion_3_vs_t
      real (C_DOUBLE), dimension (:), pointer :: Quaternion_4_vs_t

      real (C_DOUBLE), dimension (:), pointer :: AngVelocity_1_vs_t
      real (C_DOUBLE), dimension (:), pointer :: AngVelocity_2_vs_t
      real (C_DOUBLE), dimension (:), pointer :: AngVelocity_3_vs_t

      real (C_FLOAT), dimension (:), pointer :: X_position_vs_t
      real (C_FLOAT), dimension (:), pointer :: Y_position_vs_t
      real (C_FLOAT), dimension (:), pointer :: Z_position_vs_t

      real (C_FLOAT), dimension (:), pointer :: X_velocity_vs_t
      real (C_FLOAT), dimension (:), pointer :: Y_velocity_vs_t
      real (C_FLOAT), dimension (:), pointer :: Z_velocity_vs_t


      ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


      call Read_Lvl1_SC_PosHist  ( InputFileName, Num_History_Recs, METs_of_Records_Cptr,                        &
               Quaternion_1_vs_t_Cptr, Quaternion_2_vs_t_Cptr, Quaternion_3_vs_t_Cptr, Quaternion_4_vs_t_Cptr,   &
               AngVelocity_1_vs_t_Cptr, AngVelocity_2_vs_t_Cptr, AngVelocity_3_vs_t_Cptr,                        &
                   X_position_vs_t_Cptr, Y_position_vs_t_Cptr, Z_position_vs_t_Cptr,                             &
                   X_velocity_vs_t_Cptr, Y_velocity_vs_t_Cptr, Z_velocity_vs_t_Cptr )

      !  Tricky case -- variable length arrays that had memory allocated in the C subroutine
      !  using malloc -- from C we receive pointers to that memory -- now associate those
      !  pointers to Fortran arrays of the correct shape and size:


      call c_f_pointer ( METs_of_Records_Cptr, METs_of_Records, [ Num_History_Recs ] )

      call c_f_pointer ( Quaternion_1_vs_t_Cptr, Quaternion_1_vs_t, [ Num_History_Recs ] )
      call c_f_pointer ( Quaternion_2_vs_t_Cptr, Quaternion_2_vs_t, [ Num_History_Recs ] )
      call c_f_pointer ( Quaternion_3_vs_t_Cptr, Quaternion_3_vs_t, [ Num_History_Recs ] )
      call c_f_pointer ( Quaternion_4_vs_t_Cptr, Quaternion_4_vs_t, [ Num_History_Recs ] )

      call c_f_pointer ( AngVelocity_1_vs_t_Cptr, AngVelocity_1_vs_t, [ Num_History_Recs ] )
      call c_f_pointer ( AngVelocity_2_vs_t_Cptr, AngVelocity_2_vs_t, [ Num_History_Recs ] )
      call c_f_pointer ( AngVelocity_3_vs_t_Cptr, AngVelocity_3_vs_t, [ Num_History_Recs ] )

      call c_f_pointer ( X_position_vs_t_Cptr, X_position_vs_t, [ Num_History_Recs ] )
      call c_f_pointer ( Y_position_vs_t_Cptr, Y_position_vs_t, [ Num_History_Recs ] )
      call c_f_pointer ( Z_position_vs_t_Cptr, Z_position_vs_t, [ Num_History_Recs ] )

      call c_f_pointer ( X_velocity_vs_t_Cptr, X_velocity_vs_t, [ Num_History_Recs ] )
      call c_f_pointer ( Y_velocity_vs_t_Cptr, Y_velocity_vs_t, [ Num_History_Recs ] )
      call c_f_pointer ( Z_velocity_vs_t_Cptr, Z_velocity_vs_t, [ Num_History_Recs ] )

      return

   end subroutine Read_Lvl1_SC_PosHist_Fort


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


   subroutine  Cleanup_Read_Lvl1_SC_PosHist_Fort ()

      call Cleanup_Read_Lvl1_SC_PosHist  ( METs_of_Records_Cptr,                     &
              Quaternion_1_vs_t_Cptr, Quaternion_2_vs_t_Cptr, Quaternion_3_vs_t_Cptr, Quaternion_4_vs_t_Cptr,   &
              AngVelocity_1_vs_t_Cptr, AngVelocity_2_vs_t_Cptr, AngVelocity_3_vs_t_Cptr,                        &
              X_position_vs_t_Cptr, Y_position_vs_t_Cptr, Z_position_vs_t_Cptr,      &
              X_velocity_vs_t_Cptr, Y_velocity_vs_t_Cptr, Z_velocity_vs_t_Cptr )

      return

   end subroutine Cleanup_Read_Lvl1_SC_PosHist_Fort


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


   !  Interpolates between two records in the SC Position History.
   !  The "interpolation" is actually updating the position data (Cartesian Position or
   !  orientation Quaternion) with the corresponding velocity (Cartesian velocity or
   !  angular velocity).   All four items are data in the SC Position History file.
   !  Those two records that bracket the requested time have already been located and
   !  their data is passed to this subroutine.
   !  Also converts the position in Cartesian coordinates to geographic longitude & latitude,
   !  and altitude above the Earth.
   !  M. S. Briggs, 2010 May 9.    (Was previously part of the program SC_GeoCoord.)
   !  M.S.B, 2010 June 9: correct typo in output label: "km" corrected to "m".

   !  MSB, 2013 Jan 3:
   !    1) optional argument "Success: was used in one place that where the time gap was checked,
   !       but not in the other place -- fixed.
   !    2) Duration of time gap is now parameter Allowed_Gap.  Previously was 3.0 or 3.01 seconds.
   !    3) If Success = FALSE, output variables are set to special values.

   !  MSB, 2013 Dec 15:
   !    1) Now using WGS84 geodetic (ellipsoidal Earth) coordinates, unless overridden with
   !       compile-time option SPHEREEARTH.
   !    2) The quaternion is now updated with the angular velocity.
   !    3) The code to determination the time-step of the update is cleaner.
   !    4) The code now support a small "interpolation" before the first record in the
   !       SC Position History file, in addition to after the last record (which it
   !       already did.   The maximum extrapolation before or after the end of the file
   !       is AllowedPredictTime, currently 3.02 s.


   subroutine  Interpolate_Convert_SC_PosHist  ( Request_MET, Before_MET, After_MET,              &
                   Before_Quaternion, After_Quaternion, Before_AngVelocity, After_AngVelocity,    &
                   Before_Position, After_Position, Before_Velocity, After_Velocity,              &
                   Position, EastLongitude, Latitude, altitude, Quaternion, Success, FailReason )

      use Time_and_Coord_Routines

      use MSB_Constants

      real (c_double), intent (in) :: Request_MET      ! Mission Elapsed Time, s
      real (c_double), intent (in) :: Before_MET
      real (c_double), intent (in) :: After_MET

      real (c_double), dimension (4), intent (in) :: Before_Quaternion
      real (c_double), dimension (4), intent (in) :: After_Quaternion

      real (c_double), dimension (3), intent (in) :: Before_AngVelocity    !  rad / s
      real (c_double), dimension (3), intent (in) :: After_AngVelocity

      real (c_float), dimension (3), intent (in) :: Before_Position    !  m
      real (c_float), dimension (3), intent (in) :: After_Position

      real (c_float), dimension (3), intent (in) :: Before_Velocity    !  m / s
      real (c_float), dimension (3), intent (in) :: After_Velocity


      real (c_float), dimension (3), intent (out) :: Position    ! m

      real (real64), intent (out) :: EastLongitude    ! degrees
      real (real64), intent (out) :: Latitude         ! degrees
      real (real64), intent (out) :: altitude      ! m about Earth's mean radius

      real (real64), dimension (4), intent (out) :: Quaternion

      logical, intent (out) :: Success
      character (len=26), intent (out) :: FailReason


      ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


      real (c_double) :: delta_t_before, delta_t_after, Closest_delta_t
      real (c_double), dimension (3) :: Closest_Position, Closest_Velocity, Closest_AngVelocity
      real (c_double), dimension (4) :: Closest_Quaternion, AngVelocityQuaternion
      real (c_double) :: AngVelocityMagnitude, theta, s

      !  Added 2013 Jan 3.  Used fixed constants in the code before this data.
      real (c_double), parameter :: AllowedPredictTime = 3.02

      ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


      !  Bad values unless we are able to calculate the correct values:

      EastLongitude = 999.99
      Latitude = -99.99
      altitude = -99.9
      Quaternion = 0.0_real64
      Position = -999999.9_c_float

      Success = .FALSE.


      if ( Request_MET < Before_MET - AllowedPredictTime  .OR.  Request_MET > After_MET + AllowedPredictTime ) then

         write (*, *) 'ERROR: output times do not bracket the input time!'
         write (*, *) 'And the extrapolation required is too large!'
         write (*, *) 'input time:', Request_MET
         write (*, *) 'Output times:', Before_MET, After_MET
         FailReason = "SC Pos Hist data not found"

         return      !   EARLY RETURN   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      end if


      !  Stepping from the known position to the position at the time of interest
      !  is based on a very small step, of duration nomimally one second -- safe "extrapolation".
      !  Check that the above sentence is true...   That the interval between the
      !  data times is small ... we accept up to two missing records and therefore an
      !  interval of up to 3 seconds.

      if ( After_MET - Before_MET > 2.01_c_double ) then  ! gap ?

         !  The factor of two is because, if the end points are separated by 2 * AllowedPredictTime,
         !  then Request_MET will always be within AllowedPredictTime of one of them.

         if ( After_MET - Before_MET < 2.0_c_double * AllowedPredictTime ) then

            !  Gap but short enough that extrapolation by <= AllowedPredictTime
            !  is possible.

            write (*, *) 'WARNING: The difference between the bracketing times is large!'
            write (*, *)  Before_MET, After_MET

         else

            !  The difference is unacceptably large ...

            write (*, *) 'ERROR: The difference between the bracketing times is excessively large!'
            write (*, *)  Before_MET, After_MET
            FailReason = "SC Pos Hist: large gap"

            return      !   EARLY RETURN   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

         end if

      end if  ! gap ?


      Success = .TRUE.
      FailReason = "OK"

      !  Use the Position and Quaternion (orientation) from the record closest in
      !  time to the requested time.   For the Position this is a very minor improvement
      !  since both values are very accurate due to extrapolating with the velocity.

      !  Calculate the delta-times between the requested time and the bracketing times.
      !  If the bracket times successfully bracket the requested time, then
      !  delta_t_before will be positive and delta_t_after negative.   If the requested
      !  time is after the last time in the file, then both will be positive.
      !  If the requested time is before the last time in the file, both will be
      !  negative.

      delta_t_before = Request_MET - Before_MET
      delta_t_after =  Request_MET - After_MET


      !  Extrapolate from whichever bracketing time is closer.
      !  The test works for the case that the requested time is before the
      !  first time in the file or after the last time in the file.

      if ( delta_t_before > abs (delta_t_after) ) then  ! closest bracketing time ?

         Closest_delta_t = delta_t_after
         Closest_Position = After_Position
         Closest_Velocity = After_Velocity
         Closest_AngVelocity = After_AngVelocity
         Closest_Quaternion = After_Quaternion

      else  ! closest bracketing time ?

         Closest_delta_t = delta_t_before
         Closest_Position = Before_Position
         Closest_Velocity = Before_Velocity
         Closest_AngVelocity = Before_AngVelocity
         Closest_Quaternion = Before_Quaternion

      end if  ! closest bracketing time ?


      Position = real ( Closest_Position + Closest_delta_t * Closest_Velocity, c_float )

      !  Update the quaternion by the angular velocity.   See my (MSB) Quaternion Memo and also
      !  http://stackoverflow.com/questions/8816785/looking-for-a-better-method-to-do-quaternion-differentiaton

      AngVelocityMagnitude = sqrt ( dot_product ( Closest_AngVelocity, Closest_AngVelocity ) )

      if ( AngVelocityMagnitude > 1.0E-20_c_double ) then

         theta = 0.5_c_double * AngVelocityMagnitude * Closest_delta_t
         s = sinc (theta) * 0.5_c_double * Closest_delta_t
         AngVelocityQuaternion = [ s * Closest_AngVelocity (1), s * Closest_AngVelocity (2), s * Closest_AngVelocity (3),   &
                  cos (theta) ]

         Quaternion = QuaternionMultiply ( Closest_Quaternion, AngVelocityQuaternion )

      else

         ! Extremely low angular velocity.  Probably not physical for Fermi.
         !  This branch included to handle the case of an identically zero
         !  angular velocity, which will fail in the other case.

         Quaternion = Closest_Quaternion

      end if


      ! test 1:
      !theta = 0.5_c_double * sqrt ( dot_product ( Before_AngVelocity, Before_AngVelocity ) )
      !s = sinc (theta) * 0.5_c_double
      !AngVelocityQuaternion = [ s * Before_AngVelocity (1), s * Before_AngVelocity (2), s * Before_AngVelocity (3), cos (theta) ]
      !Quaternion = QuaternionMultiply ( Before_Quaternion, AngVelocityQuaternion )
      !write (*, '(/ "Before, Extrapolate from before to next, then next from file:" &
      !  &/ 4 (2X, F16.13) / 4 (2X, F16.13) / 4 (2X, F16.13) )' )   &
      !     Before_Quaternion, Quaternion, After_Quaternion


      ! test 2:
      !theta = -0.5_c_double * sqrt ( dot_product ( After_AngVelocity, After_AngVelocity ) )
      !s = -sinc (theta) * 0.5_c_double
      !AngVelocityQuaternion = [ s * After_AngVelocity (1), s * After_AngVelocity (2), s * After_AngVelocity (3), cos (theta) ]
      !Quaternion = QuaternionMultiply ( After_Quaternion, AngVelocityQuaternion )
      !write (*, '(/ "Before, Extrapolate from after to before, then next from file:" &
      !  &/ 4 (2X, F16.13) / 4 (2X, F16.13) / 4 (2X, F16.13) )' )   &
      !     Before_Quaternion, Quaternion, After_Quaternion



#ifdef SPHEREEARTH
      call Cartesian_to_SphericalGeoCoord  ( real (Request_MET, real128), Position, EastLongitude, Latitude, altitude )
#else

      call Cartesian_to_Accurate_Longitude  ( real (Request_MET, real128), Position, EastLongitude, Success )
      if ( .NOT. Success )  FailReason = "Could not calculate UT1"
      call Cartesian_to_GeoCoord_WGS84 ( Position, Latitude, altitude )

#endif


      return

   end subroutine  Interpolate_Convert_SC_PosHist


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


function QuaternionMultiply ( Q1, Q2 )

   real (c_double), dimension (4) :: QuaternionMultiply

   real (c_double), dimension (4), intent (in) :: Q1, Q2

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   QuaternionMultiply (1:3)  = Q1 (4) * Q2 (1:3)  +  Q2 (4) * Q1 (1:3)  +  CrossProd3 ( Q1 (1:3), Q2 (1:3) )

   QuaternionMultiply (4) = Q1 (4) * Q2 (4)  - dot_product ( Q1 (1:3), Q2 (1:3) )

   !  re-normalize against numerical issues:
   QuaternionMultiply = QuaternionMultiply / sqrt ( dot_product ( QuaternionMultiply, QuaternionMultiply ) )

end function QuaternionMultiply


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


function CrossProd3 ( a, b )

   real (c_double), dimension (3) :: CrossProd3

   real (c_double), dimension (3), intent (in) :: a , b

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   CrossProd3 (1) = a(2) * b(3)  -  a(3) * b(2)

   CrossProd3 (2) = a(3) * b(1)  -  a(1) * b(3)

   CrossProd3 (3) = a(1) * b(2)  -  a(2) * b(1)


end function CrossProd3


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Not defined for x = 0!
!  Use the first two terms of the Taylor series when X is small,
!  as numerically more stable.

function sinc (x)

   real (c_double) :: sinc

   real (c_double), intent (in) :: x

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   if ( x > 1.0E-6_c_double ) then

      sinc = sin (x) / x

   else

      sinc = 1.0_c_double - x**2 / 6.0_c_double

   end if


end function sinc


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


end module Read_Lvl1_SC_PosHist_mod
